from typing import List, Sequence, Union

from pydantic import BaseModel, Field, validator

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, PopulatedWithFieldnames, TemplatedVariable

from .aggregate import Aggregation, AggregationWithVariables


class TotalDimension(BaseModel):
    class Config(PopulatedWithFieldnames):
        ...

    total_column: ColumnName = Field(alias='totalColumn')
    total_rows_label: str = Field(alias='totalRowsLabel')


class TotalsStep(BaseStep):
    name = Field('totals', const=True)
    total_dimensions: List[TotalDimension] = Field(alias='totalDimensions')
    aggregations: Sequence[Aggregation]
    groups: List[ColumnName] = Field(min_items=0, default=[])

    @validator('aggregations')
    def aggregation_must_not_be_empty(cls, value):
        if len(value) < 1:
            raise ValueError('aggregations must contain at least one item')
        return value


class TotalsStepWithVariable(TotalsStep, StepWithVariablesMixin):
    aggregations: Sequence[AggregationWithVariables]
    groups: Union[TemplatedVariable, List[TemplatedVariable]]



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import ReplaceStep


def execute_replace(
    step: ReplaceStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    return df.assign(
        **{
            step.search_column: df[step.search_column].replace(
                {old_value: new_value for (old_value, new_value) in step.to_replace}
            )
        }
    )



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import TopStep


def execute_top(
    step: TopStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    pandas_method = 'nlargest' if step.sort == 'desc' else 'nsmallest'
    if step.groups:
        return (
            df.groupby(step.groups)
            .apply(lambda df: getattr(df, pandas_method)(step.limit, step.rank_on))
            .reset_index(drop=True)
        )
    else:
        return getattr(df, pandas_method)(step.limit, step.rank_on)



from typing import List, Optional, Sequence, Union

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable

from .aggregate import Aggregation, AggregationWithVariables


class RollupStep(BaseStep):
    name = Field('rollup', const=True)
    hierarchy: List[ColumnName]
    # The list of columnns to aggregate, with related aggregation function to use:
    aggregations: Sequence[Aggregation]
    # Groupby columns if rollup has to be performed by groups:
    groupby: Optional[List[ColumnName]]
    # To give a custom name to the output label column:
    label_col: Optional[ColumnName] = Field(alias='labelCol')
    # To give a custom name to the output level column:
    level_col: Optional[ColumnName] = Field(alias='levelCol')
    # To give a custom name to the output parent column:
    parent_label_col: Optional[ColumnName] = Field(alias='parentLabelCol')


class RollupStepWithVariable(RollupStep, StepWithVariablesMixin):
    aggregations: Sequence[AggregationWithVariables]
    hierarchy: Union[TemplatedVariable, List[TemplatedVariable]]
    groupby: Union[TemplatedVariable, List[TemplatedVariable]]



from typing import List, Union

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable


class UniqueGroupsStep(BaseStep):
    name = Field('uniquegroups', const=True)
    on: List[ColumnName]


class UniqueGroupsStepWithVariable(UniqueGroupsStep, StepWithVariablesMixin):
    on: Union[TemplatedVariable, List[TemplatedVariable]]



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import UnpivotStep


def execute_unpivot(
    step: UnpivotStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    df_melted = df.melt(
        id_vars=step.keep,
        value_vars=step.unpivot,
        var_name=step.unpivot_column_name,
        value_name=step.value_column_name,
    )
    return df_melted.dropna(subset=[step.value_column_name]) if step.dropna else df_melted



from pandas import Series, to_datetime, to_numeric


def cast_to_float(s: Series) -> Series:
    return to_numeric(s, errors='coerce')  # cast errors will result in NaN values


def cast_to_int(s: Series) -> Series:
    s = cast_to_float(s)
    # Since .astype(int) would fail on NaN values, treat them separately:
    s_integer = s[s.notna()].astype(int)
    s_nan = s[s.isna()]
    return s_integer.append(s_nan).sort_index()


def cast_to_str(s: Series) -> Series:
    return s.astype(str)


def cast_to_datetime(s: Series) -> Series:
    return to_datetime(s, errors='coerce')  # cast errors will result in NaT values


def cast_to_bool(s: Series) -> Series:
    return s.astype(bool)



from typing import List, Literal, Optional, Union

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable


class RankStep(BaseStep):
    name = Field('rank', const=True)
    value_col: ColumnName = Field(alias='valueCol')
    order: Literal['asc', 'desc']
    method: Literal['standard', 'dense']
    groupby: List[ColumnName] = []
    new_column_name: Optional[ColumnName] = Field(None, alias='newColumnName')


class RankStepWithVariable(RankStep, StepWithVariablesMixin):
    groupby: Union[TemplatedVariable, List[TemplatedVariable]]



from typing import List

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class DeleteStep(BaseStep):
    name = Field('delete', const=True)
    columns: List[ColumnName]



from typing import List, Literal, Union

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable


class AddMissingDatesStep(BaseStep):
    name = Field('addmissingdates', const=True)
    dates_column: ColumnName = Field(alias='datesColumn')
    dates_granularity: Union[
        Literal['day'], Literal['week'], Literal['month'], Literal['year']
    ] = Field(alias='datesGranularity')
    groups: List[ColumnName] = []


class AddMissingDatesStepWithVariables(AddMissingDatesStep, StepWithVariablesMixin):
    groups: Union[List[TemplatedVariable], TemplatedVariable]
    dates_column: TemplatedVariable = Field(alias='datesColumn')



from pandas import DataFrame, DateOffset

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.exceptions import DuplicateError
from weaverbird.pipeline.steps import EvolutionStep

OFFSETS = {
    'vsLastYear': DateOffset(years=1),
    'vsLastMonth': DateOffset(months=1),
    'vsLastWeek': DateOffset(weeks=1),
    'vsLastDay': DateOffset(days=1),
}


def execute_evolution(
    step: EvolutionStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    new_column = step.new_column or f'{step.value_col}_EVOL_{step.evolution_format.upper()}'

    id_cols = [step.date_col] + step.index_columns
    if df.set_index(id_cols).index.duplicated().any():
        raise DuplicateError('Multiple rows for the same date. Did you forget indexColumns?')

    date_col_offseted = df[step.date_col] + OFFSETS[step.evolution_type]
    df_offseted = df.assign(**{step.date_col: date_col_offseted})
    both = df.merge(df_offseted, on=id_cols, how='left', suffixes=(None, '_prev_date'))
    value_date, value_prev_date = both[step.value_col], both[f'{step.value_col}_prev_date']

    if step.evolution_format == 'abs':
        evolution = value_date - value_prev_date
    elif step.evolution_format == 'pct':
        evolution = value_date / value_prev_date - 1

    return df.assign(**{new_column: evolution})
